
from __future__ import print_function

import _setup_test_env  # noqa
import sys
import os
import threading
from returnn.util.task_system import *
import unittest
import gc

SharedMemNumpyConfig["enabled"] = True
SharedMemNumpyConfig["auto_pickling_min_size"] = 1

try:
  # noinspection PyCompatibility
  from StringIO import StringIO
except ImportError:  # Python 3
  from io import BytesIO as StringIO


def pickle_dumps(obj):
  sio = StringIO()
  p = Pickler(sio)
  p.dump(obj)
  return sio.getvalue()


def pickle_loads(s):
  p = Unpickler(StringIO(s))
  return p.load()


def find_numpy_shared_by_shmid(shmid):
  for sh in SharedNumpyArray.ServerInstances:
    assert isinstance(sh, SharedNumpyArray)
    assert sh.mem is not None
    assert sh.mem.shmid > 0
    if sh.mem.shmid == shmid:
      return sh
  return None


_have_working_shmget = None


def have_working_shmget():
  global _have_working_shmget
  if _have_working_shmget is None:
    _have_working_shmget = SharedMem.is_shmget_functioning()
  print("shmget functioning:", _have_working_shmget)
  return _have_working_shmget


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_shmget_functioning():
  assert SharedMem.is_shmget_functioning()


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_numpy():
  m = numpy.random.randn(10, 10)
  p = pickle_dumps(m)
  m2 = pickle_loads(p)
  assert isinstance(m2, numpy.ndarray)
  assert numpy.allclose(m, m2)
  assert isinstance(m2.base, SharedNumpyArray)
  shared_client = m2.base
  assert not shared_client.is_server
  shared_server = find_numpy_shared_by_shmid(m2.base.mem.shmid)
  assert shared_server.is_server
  assert numpy.allclose(m, shared_server.create_numpy_array())
  assert numpy.allclose(m, shared_client.create_numpy_array())
  assert shared_server.is_in_use()
  assert shared_client.is_in_use()
  numpy_set_unused(m2)
  assert not shared_server.is_in_use()
  assert shared_client.mem is None


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_numpy_scalar():
  # Note that a scalar really does not work because numpy.array(float) will always own its data.
  m = numpy.array([numpy.random.randn()])
  assert isinstance(m, numpy.ndarray)
  assert m.shape == (1,)
  assert m.nbytes >= 1
  p = pickle_dumps(m)
  m2 = pickle_loads(p)
  assert isinstance(m2, numpy.ndarray)
  assert numpy.allclose(m, m2)
  assert isinstance(m2.base, SharedNumpyArray)
  shared_client = m2.base
  assert not shared_client.is_server
  shared_server = find_numpy_shared_by_shmid(m2.base.mem.shmid)
  assert shared_server.is_server
  assert numpy.allclose(m, shared_server.create_numpy_array())
  assert numpy.allclose(m, shared_client.create_numpy_array())
  assert shared_server.is_in_use()
  assert shared_client.is_in_use()
  numpy_set_unused(m2)
  assert not shared_server.is_in_use()
  assert shared_client.mem is None


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_gc_aggressive():
  m = numpy.random.randn(10, 10)
  p = pickle_dumps(m)
  m2 = pickle_loads(p)
  assert isinstance(m2, numpy.ndarray)
  assert numpy.allclose(m, m2)
  assert isinstance(m2.base, SharedNumpyArray)
  print("refcount: %i" % sys.getrefcount(m2.base))
  gc.collect()
  gc.collect()
  print("refcount: %i" % sys.getrefcount(m2.base))
  assert m2.base.is_in_use()
  server = find_numpy_shared_by_shmid(m2.base.mem.shmid)
  m2 = None
  gc.collect()


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_multiple():
  for i in range(20):
    ms = [numpy.random.randn(10, 10) for i in range(i % 3 + 1)]
    p = pickle_dumps(ms)
    ms2 = pickle_loads(p)
    assert len(ms) == len(ms2)
    for m, m2 in zip(ms, ms2):
      assert numpy.allclose(m, m2)
      assert isinstance(m2.base, SharedNumpyArray)


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_unpickle_auto_unused():
  old_num_servers = None
  for i in range(10):
    m = numpy.random.randn(i * 2 + 1, i * 3 + 2)
    p = pickle_dumps((m, m, m))
    new_num_servers = len(SharedNumpyArray.ServerInstances)
    if old_num_servers is not None:
      assert old_num_servers == new_num_servers
    old_num_servers = new_num_servers
    m2, m3, m4 = pickle_loads(p)
    assert numpy.allclose(m, m2)
    assert numpy.allclose(m, m3)
    assert numpy.allclose(m, m4)
    assert not m4.base.is_server
    m4.base._get_in_use_flag_ref().value = 42
    assert m4.base._get_in_use_flag_ref().value == 42
    assert find_numpy_shared_by_shmid(m4.base.mem.shmid)._get_in_use_flag_ref().value == 42
    assert numpy.allclose(m, m4)
    ss = list([find_numpy_shared_by_shmid(_m.base.mem.shmid) for _m in (m2, m3, m4)])
    _m = None
    m2 = m3 = m4 = None
    gc.collect()
    for s in ss:
      assert isinstance(s, SharedNumpyArray)
      assert s.is_server
      assert not s.is_in_use()



#!/usr/bin/env python3

"""
This is a wrapper around :func:`returnn.tf.engine.Engine.cleanup_old_models`.
"""

from __future__ import print_function
import sys
import os
import argparse

import _setup_returnn_env  # noqa
from returnn.util import better_exchook
from returnn.log import log
from returnn.__main__ import init, finalize


def main():
  """
  Main entry.
  """
  arg_parser = argparse.ArgumentParser()
  arg_parser.add_argument("--config")
  arg_parser.add_argument("--cwd", help="will change to this dir")
  arg_parser.add_argument("--model", help="model filenames")
  arg_parser.add_argument("--scores", help="learning_rate_control file, e.g. newbob.data")
  arg_parser.add_argument("--dry_run", action="store_true")
  args = arg_parser.parse_args()
  return_code = 0
  try:
    if args.cwd:
      os.chdir(args.cwd)
    init(
      extra_greeting="Delete old models.",
      config_filename=args.config or None,
      config_updates={
        "use_tensorflow": True,
        "need_data": False,
        "device": "cpu"})
    from returnn.__main__ import engine, config
    if args.model:
      config.set("model", args.model)
    if args.scores:
      config.set("learning_rate_file", args.scores)
    if args.dry_run:
      config.set("dry_run", True)
    engine.cleanup_old_models(ask_for_confirmation=True)

  except KeyboardInterrupt:
    return_code = 1
    print("KeyboardInterrupt", file=getattr(log, "v3", sys.stderr))
    if getattr(log, "verbose", [False] * 6)[5]:
      sys.excepthook(*sys.exc_info())
  finalize()
  if return_code:
    sys.exit(return_code)


if __name__ == "__main__":
  better_exchook.install()
  main()





import os
import sys

os.environ["THEANO_FLAGS"] = "mode=FAST_COMPILE"
import _setup_test_env  # noqa
from returnn.theano.layers.ctc import *
from nose.tools import assert_equal


def test1_seq():
  pcx = numpy.array([[0.1, 0.3, 0.2, 0.1, 0.3],
                     [0.3, 0.1, 0.2, 0.1, 0.3],
                     [0.2, 0.1, 0.4, 0.2, 0.1]])
  assert numpy.allclose(pcx.sum(axis=1), 1)
  log_pcx = numpy.log(pcx)
  targets = numpy.array([1,2])
  return log_pcx, targets


def manual_ctc(log_pcx, targets):
  # pcx is (time,label)
  # targets is (seqlen,)
  pcx = numpy.exp(log_pcx)
  T, L = pcx.shape
  S, = targets.shape
  pcx_y = pcx[:, targets]  # (time,seqlen)
  fw = numpy.zeros_like(pcx_y)  # (time,seqlen)
  for t in range(T):
    if t > 0:
      fw_prev = fw[t - 1]
    else:
      fw_prev = numpy.array([1.] + [0.] * (S - 1))  # initial
    fw[t] = fw_prev  # loops
    fw[t, 1:] += fw_prev[:-1]  # forward step
    fw[t] *= pcx_y[t]
  bw = numpy.zeros_like(pcx_y)  # (time,seqlen)
  for t in reversed(range(T)):
    if t < T - 1:
      bw_prev = bw[t + 1]
    else:
      bw_prev = numpy.array([0.] * (S - 1) + [1.])  # initial
    bw[t] = bw_prev  # loops
    bw[t, :-1] += bw_prev[1:]  # backward step
    bw[t] *= pcx_y[t]
  probs = (fw * bw) / pcx_y
  return probs

def make_batch(seqs):
  assert seqs
  max_time = 0
  max_seq_len = 0
  label_dim = None
  for log_pcx, targets in seqs:
    assert len(log_pcx.shape) == 2
    assert len(targets.shape) == 1
    if label_dim is None: label_dim = log_pcx.shape[1]
    assert label_dim == log_pcx.shape[1]
    max_time = max(max_time, log_pcx.shape[0])
    max_seq_len = max(max_seq_len, targets.shape[0])
  batch_log_pcx = numpy.zeros((max_time, len(seqs), label_dim), dtype='float32')
  batch_targets = numpy.zeros((max_seq_len, len(seqs)), dtype='int32')
  time_mask = numpy.zeros((max_time, len(seqs)), dtype='int8')
  for batch_idx, (log_pcx, targets) in enumerate(seqs):
    batch_log_pcx[:,batch_idx,:] = log_pcx
    batch_targets[:,batch_idx] = targets
    time_mask[0:len(log_pcx), batch_idx] = 1
  seq_lens = numpy.array([len(targets) for _, targets in seqs], dtype='int32')
  return batch_log_pcx, time_mask, batch_targets, seq_lens



class TestNetworkCtcLayer(object):

  def test_uniq_with_lengths(self):
    s1 = [[5, 3, 3, 1, 0, 0],
          [4, 4, 1, 1, 2, 2],
          [7, 7, 2, 0, 0, 0]]
    s1 = theano.shared(numpy.array(s1).T)
    s2 = T.neq(s1, 0)
    out_seqs, seq_lens = uniq_with_lengths(s1, s2)
    out_seqs = out_seqs.eval()
    seq_lens = seq_lens.eval()
    assert_equal(list(seq_lens), [3, 3, 2])
    assert_equal(out_seqs.shape, (3, 3))
    assert_equal(list(out_seqs[:, 0]), [5, 3, 1])
    assert_equal(list(out_seqs[:, 1]), [4, 1, 2])
    assert_equal(list(out_seqs[:, 2])[:2], [7, 2])

  def test_ctc(self):
    log_pcx, time_mask, targets, seq_lens = map(theano.shared, make_batch([test1_seq()]))
    theano_log_ctc_probs = ctc(log_pcx, time_mask, targets, seq_lens)
    manual_ctc_probs = manual_ctc(*test1_seq())
    assert numpy.allclose(numpy.exp(theano_log_ctc_probs.eval())[:,0,:], manual_ctc_probs)
    theano_ctc_total = log_sum(theano_log_ctc_probs, axis=2)
    manual_ctc_total = manual_ctc_probs.sum(axis=1)
    assert numpy.allclose(numpy.exp(theano_ctc_total.eval()[:,0]), manual_ctc_total)
